import { upgradeActions } from './upgrade/actions.js';

const upgradeMessage = 'upgradeMessage';
const successClass = 'iw-positive';
const failClass = 'iw-negative';
const payloadStatusClass = 'iw-status';
const runUpgradeBtnId = 'run-upgrade-btn';
const hiddenClass = 'invisible';
const payloadClass = 'iw-payload';
const payloadContainerId = 'iw-payload-container';
const upgradeInProgressContainerId = 'upgrade-in-progress-container';
const upgradeStartedId = 'upgrade-started';
const upgradeInProgressId = 'upgrade-in-progress';
const upgradeDoneId = 'upgrade-done';
const spinnerClass = 'list-spinner';
const submitButtonId = 'start_upgrade-submit';
const cancelButtonClass = 'iw-form-cancel';


let registerUpgradeEventListenerAndDisplayMessages = () => {
    displayBannerMessage();
    updatePageContentOnDomChange();
}

/**
 * Display banner message after page refresh.
 *
 * In case upgrade fails, this should be set
 * in session storage in the fail function.
 */
let displayBannerMessage = () => {
    let message = getSessionStorageMessage();

    if (message) {
        iw_lib.iw_message_banner(message.message, message.className);
    }

    // remove message from session storage so
    // that it won't show up next time the page
    // is refreshed.
    sessionStorage.removeItem(upgradeMessage)
}

let getSessionStorageMessage = () => {
    let message = sessionStorage.getItem(upgradeMessage)
    if (message) {
        return JSON.parse(message);
    }
}

/*
 * Set message, e.g. if upgrade fails.
 */
let setMessage = (msg, className) => sessionStorage.setItem(upgradeMessage, JSON.stringify({
    message: msg,
    className: className
}));

/**
 * Detect if nodes change and update page content
 * if status of payload changes.
 */
let updatePageContentOnDomChange = () => {
    let container = $('#upgrade-major-version-container').get(0);
    let observer = new MutationObserver(updatePageContent);
    let config = { attributes: true, childList: true, subtree: true }
    observer.observe(container, config)
}

let updatePageContent = () => {
    let payloadExists = $('#iw-payload-container').find(`.${payloadClass}`).length > 1;

    if (payloadExists) {
        showPayloadMessage();
        showUpdateLinkIfNoErrors();
    }
}

/**
 * Hidden message elements for opt in and opt out are included in the template.
 *
 * Remove class that hides the element that we wish to display,
 * based on opt in status in payload.
 */
let showPayloadMessage = () => {
    let $statusElements = $(`.${payloadStatusClass}`);
    let $isOptedInStatusElement = $($statusElements[0]);
    let isOptedIn = $isOptedInStatusElement.hasClass(successClass);
    let $msgElement = isOptedIn ? $('#iw-upgrade-optedin') : $('#iw-upgrade-optedout');
    $msgElement.removeClass(hiddenClass);
}

let showUpdateLinkIfNoErrors = () => {
    let $runUpgradeBtn = $(`#${runUpgradeBtnId}`);
    let btnDisabled = $runUpgradeBtn.prop('disabled');
    let noErrors = $(`.${payloadStatusClass}.${failClass}`).length === 0;

    if (noErrors && btnDisabled) {
        $runUpgradeBtn.prop('disabled', false);
        $runUpgradeBtn.off('click', displayUpgradeForm);
        $runUpgradeBtn.on('click', displayUpgradeForm);
    }
}

let displayUpgradeForm = () => {
    upgradeActions.getUpgradeFormHtml()
        .done(data => displayForm(data))
        .fail((xhr) => displayFailure(xhr))
}

let displayForm = iworxData => {
    let res = getParsedResponse(iworxData);

    let formHtml = res.data.html;
    let labelSpan = $(formHtml).find('.iw-form-label')[0];
    let header = $(labelSpan).text();
    let insertBreakpoints = false;

    iw_lib.popupModal(formHtml, header, insertBreakpoints);
    registerUpgradeFormEventHandlers();
}

let registerUpgradeFormEventHandlers = () => {
    let $startUpgradeButton = $(`#${submitButtonId}`);
    let $cancelButton = $($(`.${cancelButtonClass}`)[0]);

    // prevent multiple events of same sort
    $startUpgradeButton.off('click', startUpgrade);
    $startUpgradeButton.on('click', startUpgrade);

    $cancelButton.off('click', closeModal);
    $cancelButton.on('click', closeModal);
}

let startUpgrade = event => {
    event.preventDefault();

    let softwareChannel = $(`option:selected`).text();

    iw_lib.iw_modal_popup.hide();
    renderBackupStarted();

    upgradeActions.startUpgrade(softwareChannel)
        .done(data => handleIwUpgradeFailure(data))
        .fail(xhr => handleExpectedResponseFailure(xhr));
}

let closeModal =  () => iw_lib.iw_modal_popup.hide();

let renderBackupStarted = () => {
    // hide payload & display in progress content
    $(`#${payloadContainerId}`).addClass(hiddenClass);
    $(`#${upgradeInProgressContainerId}`).removeClass(hiddenClass);

    // display first step - starting upgrade
    $(`#${upgradeStartedId}`).removeClass(hiddenClass);
    $(`#${upgradeStartedId}`).addClass(spinnerClass);
}

/**
 * Should only get here if upgrade fails.
 *
 * The xhr request will not get an answer if
 * upgrade started, since the server goes down
 * during upgrade.
 */
let handleIwUpgradeFailure = (iwResponseData) => {
    let res = getParsedResponse(iwResponseData);

    if (res.code !== 201) {
        let backendErrorMessage = 'Upgrade failed. Please open a support ticket.';

        // response object will not contain
        // data & message if backend code error
        setMessage(res.data && res.data.message || backendErrorMessage, failClass);
        window.location.reload();
    } else {
        // should never really get here as upgrade will
        // interrupt response when it upgrades & restarts the server
        console.log('got ok when shouldnt')
        console.log(iwResponseData)
    }
}

/**
 * Assume that server is down/in maintanence mode
 * if it failed to respond.
 */
let handleExpectedResponseFailure = () => {
    renderInProgress();
    awaitUpgrade();
}

let renderInProgress = () => {
    $(`#${upgradeStartedId}`).removeClass(spinnerClass);

    $(`#${upgradeInProgressId}`).removeClass(hiddenClass);
    $(`#${upgradeInProgressId}`).addClass(spinnerClass);
}


let awaitUpgrade = () => setTimeout(checkVersion, 5000)
let checkVersion = () => upgradeActions.getIwVersion()
    .done(data => handleIwVersionResponseSuccess(data))
    .fail(awaitUpgrade); //assume server is still in maintenence mode

let handleIwVersionResponseSuccess = data => {
    let res = getParsedResponse(data);
    let version = res.data.version && res.data.version[0];

    if (res.code !== 201 || version != 7) {
        // TODO display error message
        // though shouldn't get here...
        console.log('Something is wrong...');
        console.log(res)
    } else {
        // assume upgrade done

        $(`#${upgradeInProgressId}`).removeClass(spinnerClass);
        $(`#${upgradeDoneId}`).removeClass(hiddenClass);

        setTimeout(() => {
            window.location.href = '/nodeworx/welcome';
        }, 5000);
    }

}


let displayFailure = (xhr) => {
    console.log('SOMETHING IS WRONG')
}

let getParsedResponse = (response) => {
    let res = JSON.parse(response);

    return {
        code: res.c,
        data: res.d,
        text: res.t
    };
}


registerUpgradeEventListenerAndDisplayMessages();