<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @subpackage API
 * @author     Paul Oehler <poehler@interworx.com>
 * @date       Nov 22, 2008
 */

/**
 * Clustering related api functions.
 *
 * @package    InterWorx
 * @subpackage API
 */
class IWorx_API_Cluster
{

    /**
     * Checks to see if this server is in a cluster.
     *
     * @param array|struct $params
     * @return integer
     */
    public function is_incluster($params)
    {
        API::auth($params);
        return (Node::isInCluster()) ? 1 : 0;
    }

    /**
     * Checks to see if this server is a "realserver".
     *
     * @param array|struct $params
     * @return integer
     */
    public function is_realserver($params)
    {
        API::auth($params);
        return (Node::isRealServer()) ? 1 : 0;
    }

    /**
     * Checks to see if this server is a cluster director.
     *
     * @param array|struct $params
     * @return integer
     */
    public function is_director($params)
    {
        API::auth($params);
        return (Node::isDirector()) ? 1 : 0;
    }

    /**
     * Checks to see if this server is a manager.
     *
     * @param array|struct $params
     * @return integer
     */
    public function is_manager($params)
    {
        API::auth($params);
        return (Node::isManager()) ? 1 : 0;
    }

    /**
     * Checks to see if this server has the necessary free SW uids and guids.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function has_free_ids($params)
    {
        API::auth($params);
        $result = [];
        foreach ($params as $user) {
            if (isset($user['uid'])) {
                if (ShellAccount::uidExists($user['uid'])) {
                    $result['uid'][] = $user['uid'];
                }
                if (ShellAccount::gidExists($user['gid'])) {
                    $result['gid'][] = $user['gid'];
                }
            }
        }

        return $result;
    }

    /**
     * Mounts a NFS partition.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function nfs_mount($params)
    {
        API::auth($params, API::AUTH_RETURN_SESSION);

        $input = array( 'host'             => $params['ip'],
                    'remote_directory' => $params['remote'],
                    'mount_point'      => $params['local']
                    );

        foreach ($input as $key => $value) {
            IW::Env()->addInput($key, $value);
        }
        $controller_name = Ctrl_Util::convertPathToClassName('nodeworx/nfs/mount');
        $Reply = IW::FC()->route($controller_name, 'add');
        return $Reply;
    }

    /**
     * Unmounts a NFS partition.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function nfs_unmount($params)
    {
        API::auth($params, API::AUTH_RETURN_SESSION);

        IW::Env()->addInput('mounts', $params['local']);

        $controller_name = Ctrl_Util::convertPathToClassName('nodeworx/nfs/mount');
        return IW::FC()->route($controller_name, 'delete');
    }

    /**
     * Exports a dir for NFS mountage.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function nfs_export($params)
    {
        API::auth($params, API::AUTH_RETURN_SESSION);
        $rw   = '';
        $root = '';
        $sync = '';

        foreach ($params['opts'] as $opt) {
            if (array_key_exists($opt, NFS::EXPORT_READ_OPTIONS()) === true) {
                $rw = $opt;
            }
            if (array_key_exists($opt, NFS::EXPORT_ROOT_OPTIONS()) === true) {
                $root = $opt;
            }
            if (array_key_exists($opt, NFS::EXPORT_SYNC_OPTIONS()) === true) {
                $sync = $opt;
            }
        }

        $input = array( 'ip_allow_list'       => $params['ip'],
                    'directory_to_export' => $params['dir'],
                    'read_write_access'   => $rw,
                    'root_user_access'    => $root,
                    'sync_options'        => $sync );

        foreach ($input as $key => $value) {
            IW::Env()->addInput($key, $value);
        }
        $controller_name = Ctrl_Util::convertPathToClassName('nodeworx/nfs/export');
        return IW::FC()->route($controller_name, 'add');
    }


    /**
     * Removes an NFS exported dir.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function nfs_unexport($params)
    {
        API::auth($params, API::AUTH_RETURN_SESSION);

        IW::Env()->addInput('exports', array( "{$params['dir']}:{$params['ip']}" ));

        $controller_name = Ctrl_Util::convertPathToClassName('nodeworx/nfs/export');
        return IW::FC()->route($controller_name, 'delete');
    }

    /**
     * Sets a cluster node role.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function set_node_role($params)
    {
        API::auth($params);

        if (!Node::isValidRole($params['role'])) {
            return API::formatResponse(API::ERROR_INVALID_ROLE);
        }

        if (!Node::setRole($params['role'])) {
            return API::formatResponse(API::ERROR_INI_UPDATE_FAILED);
        }

        $ip = false;
        if ($params['manager'] != false) {
            if (Node::isInCluster() && !Node::isManager()) {
                if (!IWorxValidate::ipv4($params['manager'])) {
                    return API::formatResponse(API::ERROR_INVALID_IP);
                }

                $ip = $params['manager'];
            }
        }

        if (!Node::setManager($ip)) {
            return API::formatResponse(API::ERROR_INI_UPDATE_FAILED);
        }

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Sets cluster node ID.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function set_node_id($params)
    {
        API::auth($params);

        if (Node::setNodeID($params['id'])) {
            return API::formatResponse(API::SUCCESS);
        }

        return API::formatResponse(API::ERROR_INI_UPDATE_FAILED);
    }

    /**
     * Sets this server's DSN value.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function set_dsn($params)
    {
        API::auth($params);

        if (empty($params['section']) ||
        empty($params['name']) ||
        empty($params['dsn']) ) {
            return API::formatResponse(API::ERROR_INVALID_ARGS);
        }

        $section = $params['section'];
        $name    = $params['name'];
        $dsn     = $params['dsn'];

        // make a backup of our shizzle

        // NOTE: I'M GOING TO FORCE THE SAVING OF THE DSN
        // THAT HAVE SOCKETS SO WE ALWAYS HAVE THE OLD
        // SOCKET BASED SHIT AVAIALBE.  THIS WILL MEAN ONLY
        // ALLOWING THE 1ST BACKUP TO SUCCEED WHICH SHOULD
        // BE FINE FOR NOW.
        if (Ini::is_set($section, $name . '.BACKUP') === false) {
            IW::getLog()->debug("saving backup dsn INI [$section][$name.BACKUP] = $dsn");
            Ini::set($section, $name . '.BACKUP', Ini::get($section, $name));
        }

        // 2nd backup that is going to be authoritative.
        if (Ini::is_set($section, $name . '.orig') === false) {
            IW::getLog()->debug("saving ORIG dsn INI [$section][$name.orig] = $dsn");

            Ini::set($section, $name . '.orig', Ini::get($section, $name));
        }

        // set the new one
        IW::getLog()->info("updating INI [$section][$name] = $dsn");
        Ini::set($section, $name, $dsn);

        switch ($section) {
            case 'vpopmail':
                $ok = Vpopmail::updateSQLFile($dsn);
                break;

            case 'proftpd':
                $ok = ProFTP::updateSQLFile($dsn);
                break;

            case 'vpopmail.spam':
                $ok = IW::DM()->getSpamAssassin()->updateSQLFile($dsn);
                break;

            default:
                $ok = true;
        }

        Ini::write();

        if (!$ok) {
            return API::formatResponse(API::ERROR_SQL_FILE_UPDATE_FAILED);
        }

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Reverts a DSN.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function revert_dsn($params)
    {
        API::auth($params);

        $section = $params['section'];
        $name    = $params['name'];

        if (empty($section) || empty($name)) {
            return API::formatResponse(API::ERROR_INVALID_ARGS);
        }

        if (!Ini::is_set($section, $name . '.BACKUP')) {
            IW::getLog()->debug("no backup DSN found for [$section][$name]");
            return API::formatResponse(API::ERROR_NO_BACKUP_DSN);
        }

        $dsn = Ini::get($section, $name . '.BACKUP');

        IW::getLog()->info("Setting Ini::set( $section, $name, $dsn )");

        Ini::set($section, $name, $dsn);

        switch ($section) {
            case 'vpopmail':
                IW::getLog()->debug("cascading sql dsn revert to file: $section");
                $ok = Vpopmail::updateSQLFile($dsn);
                break;

            case 'proftpd':
                IW::getLog()->debug("cascading sql dsn revert to file: $section");
                $ok = ProFTP::updateSQLFile($dsn);
                break;

            case 'vpopmail.spam':
                IW::getLog()->debug("cascading sql dsn revert to file: $section");
                $ok = IW::DM()->getSpamAssassin()->updateSQLFile($dsn);
                break;

            default:
                IW::getLog()->debug("no sql file to cascade revert to for: $section");
                $ok = true;
                break;
        }//end switch

        Ini::write();

        if (!$ok) {
            IW::getLog()->debug("dsn revert failed: $section");
            return API::formatResponse(API::ERROR_SQL_FILE_UPDATE_FAILED);
        }

        IW::getLog()->debug("dsn revert success: $section");
        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Sync's IP data according to cluster manager.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function sync_ipdata($params)
    {
        $NodeWorx = API::auth($params, API::AUTH_RETURN_INTERWORX);
        Node::syncIPData();

        if ($NodeWorx->getId() <= 0) {
            IW::getLog()->debug('no such NodeWorx user for ip sync');
            return API::formatResponse(API::ERROR_INVALID_NODEWORX_ACCOUNT);
        }

        $driver = $NodeWorx->getHTTPDriver();
        $driver->syncIPData();
        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Sync the Vhosts for a domain.
     *
     * @param mixed $params the xmlrpc params array
     * @return xmlrpcresp
     */
    public function syncvhosts($params)
    {
        $NodeWorx = API::auth($params, API::AUTH_RETURN_INTERWORX);

        if ($NodeWorx->getId() <= 0) {
            IW::getLog()->debug('no such NodeWorx user for Vhosts Sync');
            return API::formatResponse(API::ERROR_INVALID_NODEWORX_ACCOUNT);
        }

        if (!isset($params['domain'])) {
            return API::formatResponse(API::ERROR_INVALID_SITEWORX_ACCOUNT);
        }

        $domain = $params['domain'];

        // get the siteworx account in question
        $SiteWorx = $NodeWorx->getSiteWorx($domain);

        if ($SiteWorx === false || $SiteWorx->getId() <= 0) {
            IW::getLog()->debug("no such SiteWorx: $domain Vhosts Sync");
            return API::formatResponse(API::ERROR_INVALID_SITEWORX_ACCOUNT);
        }

        if (($ok = $SiteWorx->syncVhosts()) !== true) {
            $codes = IWorxError::extractErrorCodes($ok);

            if (isset($codes[0])) {
                return API::formatResponse($codes[0]);
            }

            IW::getLog()->error("unknown error whiling syncing vhost : $domain");

            return API::formatResponse(API::GENERAL_FAILURE);
        }

        // just assuming this worked for now.  Maybe this is bad.
        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Bulk bind vips.
     *
     * @param mixed $params
     * @return array|struct
     */
    public function bulk_bind_vips($params)
    {
        API::auth($params);
        $ip_info = $params['ip_info'];
        foreach ($ip_info as $ip_array) {
            $ip       = $ip_array['ipaddr'];
            $type     = $ip_array['type'];
            $reseller = $ip_array['reseller'];
            $dnsip    = $ip_array['dnsip'];
            $pfm      = $ip_array['pfm'];

            IW::getLog()->info("attempting to bind VIP: $ip");

            if (!IWorxValidate::ip($ip)) {
                IW::getLog()->debug("invalid VIP to bind: $ip");
                return API::formatResponse(API::ERROR_INVALID_VIP);
            }

            if (!Node::bindVIP($ip, $type, $reseller, $dnsip, $pfm)) {
                IW::getLog()->debug("VIP bind failure: $ip");
                return API::formatResponse(API::ERROR_VIP_BIND_FAILED);
            }

            IW::getLog()->debug("bind VIP success: $ip");
        }//end foreach

        // restart all the services to re-bind the
        // New IP to the system
        IW::getLog()->debug('restarting apache to re-bind IPs');
        IW::DM()->getHttp()->restart($block = false, $failsafe = false);

        IW::getLog()->debug('restarting proftpd to re-bind IPs');
        IW::DM()->getFtp()->restart($block = false);

        IW::getLog()->debug('restarting mysql to re-bind IPs');
        IW::DM()->getDb()->restart($block = false);

        IW::getLog()->debug('restarting imap/pop3 to re-bind IPs');
        IW::DM()->getImail()->restart($block = false);

        IW::getLog()->debug('restarting qmail to re-bind IPs');
        IW::DM()->getOmail()->restart($block = false);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Binds a cluster VIP to this box.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function bind_vip($params)
    {
        API::auth($params);

        $ip    = $params['ipaddr'];
        $dnsip = $params['dnsip'];

        if (isset($params['type'])) {
            $type = $params['type'];
        } else {
            $type = IWorx_IP::SHAREDIP;
        }
        if (isset($params['reseller'])) {
            $reseller = $params['reseller'];
        } else {
            $reseller = 'no';
        }

        if (isset($params['pfm'])) {
            $pfm = $params['pfm'];
        } else {
            $pfm = LoadBalancer::LVS_DR;
        }

        IW::getLog()->info("attempting to bind VIP: $ip");

        if (!IWorxValidate::ip($ip)) {
            IW::getLog()->debug("invalid VIP to bind: $ip");
            return API::formatResponse(API::ERROR_INVALID_VIP);
        }

        if (!Node::bindVIP($ip, $type, $reseller, $dnsip, $pfm)) {
            IW::getLog()->debug("VIP bind failure: $ip");
            return API::formatResponse(API::ERROR_VIP_BIND_FAILED);
        }

        // restart all the services to re-bind the
        // New IP to the system
        IW::getLog()->debug('restarting apache to re-bind IPs');
        IW::DM()->getHttp()->restart($block = false, $failsafe = false);

        IW::getLog()->debug('restarting proftpd to re-bind IPs');
        IW::DM()->getFtp()->restart($block = false);

        IW::getLog()->debug('restarting mysql to re-bind IPs');
        IW::DM()->getDb()->restart($block = false);

        IW::getLog()->debug('restarting imap/pop3 to re-bind IPs');
        IW::DM()->getImail()->restart($block = false);

        IW::getLog()->debug('restarting qmail to re-bind IPs');
        IW::DM()->getOmail()->restart($block = false);

        IW::getLog()->debug("bind VIP success: $ip");

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Bulk bind vips.
     *
     * @param mixed $params
     * @return array|struct
     */
    public function bulk_unbind_vips($params)
    {
        API::auth($params);
        $ip_info = $params['ip_info'];

        foreach ($ip_info as $ip_array) {
            $ip = $ip_array['ipaddr'];

            IW::getLog()->info("attempting to UNbind VIP: $ip");

            $IWorx_IP = IWorx_IP::factoryFromDatabase($ip);

            if ($IWorx_IP === false) {
                IW::getLog()->debug("non-existant ip address $ip");
                return API::formatResponse(API::ERROR_VIP_UNBIND_FAILED);
            }

            if ($IWorx_IP->isActive() === false) {
                IW::getLog()->debug("VIP not active: $ip");
                return API::formatResponse(API::SUCCESS);
            }

            if (!Node::unbindVIP($ip)) {
                IW::getLog()->debug("unbind VIP failure: $ip");
                return API::formatResponse(API::ERROR_VIP_UNBIND_FAILED);
            }

            IW::getLog()->debug("unbind VIP success: $ip");
        }//end foreach

        // restart all the services to un-bind the
        // IP from the system
        IW::getLog()->debug('restarting apache to un-bind IPs');
        IW::DM()->getHttp()->restart($block = false, $failsafe = false);

        IW::getLog()->debug('restarting proftpd to un-bind IPs');
        IW::DM()->getFtp()->restart($block = false);

        IW::getLog()->debug('restarting mysql to un-bind IPs');
        IW::DM()->getDb()->restart($block = false);

        IW::getLog()->debug('restarting imap/pop3 to un-bind IPs');
        IW::DM()->getImail()->restart($block = false);

        IW::getLog()->debug('restarting qmail to un-bind IPs');
        IW::DM()->getOmail()->restart($block = false);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Unbinds a cluster VIP.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function unbind_vip($params)
    {
        API::auth($params);
        $ip = $params['ipaddr'];

        IW::getLog()->info("attempting to UNbind VIP: $ip");

        $IWorx_IP = IWorx_IP::factoryFromDatabase($ip);

        if ($IWorx_IP === false) {
            IW::getLog()->debug("non-existant ip address given {$ip}");
            return API::formatResponse(API::ERROR_VIP_UNBIND_FAILED);
        }

        if ($IWorx_IP->isActive() === false) {
            IW::getLog()->debug("VIP not active: $ip");
            return API::formatResponse(API::SUCCESS);
        }

        if (!Node::unbindVIP($ip)) {
            IW::getLog()->debug("unbind VIP failure: $ip");
            return API::formatResponse(API::ERROR_VIP_UNBIND_FAILED);
        }

        IW::getLog()->debug("unbind VIP success: $ip");

        // restart all the services to un-bind the
        // IP from the system
        IW::getLog()->debug('restarting apache to un-bind IPs');
        IW::DM()->getHttp()->restart($block = false, $failsafe = false);

        IW::getLog()->debug('restarting proftpd to un-bind IPs');
        IW::DM()->getFtp()->restart($block = false);

        IW::getLog()->debug('restarting mysql to un-bind IPs');
        IW::DM()->getDb()->restart($block = false);

        IW::getLog()->debug('restarting imap/pop3 to un-bind IPs');
        IW::DM()->getImail()->restart($block = false);

        IW::getLog()->debug('restarting qmail to un-bind IPs');
        IW::DM()->getOmail()->restart($block = false);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Restarts the webserver.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function restart_httpd($params)
    {
        API::auth($params);
        IW::getLog()->debug('restarting apache to un-bind IPs');
        IW::DM()->getHttp()->restart($block = false, $failsafe = false);
        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Sync's VIP dns service.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function sync_vip_dns_service($params)
    {
        API::auth($params);

        $cm_servers = $params['servers'];

        if (!is_array($cm_servers)) {
            //bail
            IW::getLog()->error("SYNC FAILED, CM_SERVERS ISN'T AN ARRAY");
            return API::formatResponse(API::SUCCESS);
        }

        Node::syncVipDnsService($cm_servers);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Update simscan.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function update_simscan($params)
    {
        API::auth($params);

        $line = $params['line'];

        // :clam=no,spam=no,trophie=no,spam_hits=95

        $line = str_replace(':', '', $line);
        $bits = explode(',', $line);

        $simscan = new SimScan();

        foreach ($bits as $bit) {
            list( $key, $val ) = explode('=', $bit);

            if ($key == 'spam') {
                if ($val == 'yes') {
                    $simscan->enableGlobalSpam(true);
                } else {
                    $simscan->disableGlobalSpam(true);
                }
            }
            if ($key == 'clam') {
                if ($val == 'yes') {
                    $simscan->enableGlobalClam(true);
                } else {
                    $simscan->disableGlobalClam(true);
                }
            }
            if ($key == 'spam_hits') {
                $simscan->setGlobalSpamHits($val, true);
            }
        }//end foreach

        if ($simscan->write()) {
            return API::formatResponse(API::SUCCESS);
        }

        return API::formatResponse(API::ERROR_INI_UPDATE_FAILED);
    }

    /**
     * Add alias domain.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function add_alias_domain($params)
    {
        /* @var $NodeWorx NodeWorx */
        $NodeWorx = API::auth($params, API::AUTH_RETURN_INTERWORX);

        $domain      = $params['domain'];
        $aliasdomain = $params['aliasdomain'];

        $siteworx = $NodeWorx->getSiteWorx($domain);

        if ($siteworx === false || $siteworx->getId() < 0) {
            return API::formatResponse(API::ERROR_DOMAIN_NOT_FOUND);
        }

        $imail = $siteworx->getIncomingMailDriver();
        $imail->addAliasDomain($aliasdomain);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Delete alias domain.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function delete_alias_domain($params)
    {
        /* @var $NodeWorx NodeWorx */
        $NodeWorx = API::auth($params, API::AUTH_RETURN_INTERWORX);

        $domain      = $params['domain'];
        $aliasdomain = $params['aliasdomain'];

        $siteworx = $NodeWorx->getSiteWorx($domain);

        if ($siteworx === false || $siteworx->getId() < 0) {
            return API::formatResponse(API::ERROR_DOMAIN_NOT_FOUND);
        }

        $imail = $siteworx->getIncomingMailDriver();
        $imail->deleteAliasDomain($aliasdomain);

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Get vips.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function get_vips($params)
    {
        API::auth($params);

        IW::getLog()->debug('getting VIP list');

        $ips = IWorx_IP::getIPv4Vips();

        // return a comma seperated list since we
        // can't return the array itself.
        $list = implode(',', array_keys($ips));

        return $list;
    }
}
