<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @subpackage API
 * @author     Paul Oehler <poehler@interworx.com>
 * @date       Nov 24, 2008
 */

/**
 * Dns synchronization related API functions.
 *
 * @package    InterWorx
 * @subpackage API
 */
class IWorx_API_DnsSync
{
    /**
     * Add a dns host to this box to grab or give external dns data.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function dns_add_host($params)
    {
        API::auth($params);

        IW::getLog()->info($params);

        $ip     = $params['ip'];
        $role   = $params['role'];
        $apikey = $params['sync-apikey'];

        IW::getLog()->debug("adding a DNS host: $ip");

        $host = new Dns_SyncHost();

        $host->setHostname($ip);
        $host->setRole($role);

        $host->setAPIKey($apikey);

        if (($ok = $host->validate()) !== true) {
            IW::getLog()->error("DNS host validate error: $ip");
            IW::getLog()->error(implode("\n", IWorxError::extractErrorMessages($ok)));
            $codes = IWorxError::extractErrorCodes($ok);
            $code  = ($codes[0] > 0) ? $codes[0] : API::ERROR_DNSSYNC_VALIDATION;
            return API::formatResponse($code);
        }

        $shmanager = new Dns_SyncHostManager();

        // WE DON'T WANT TO CASCADE HERE SINCE THE CASCADING
        // ONLY TAKES PLACE OUTSIDE OF THE API.  ONCE WE'RE
        // IN THE API WE NEED TO NOT CASCADE AS THIS WOULD
        // CAUSE AN INFINITE LOOP.
        if (($ok = $shmanager->addHost($host, $cascade = false)) !== true) {
            IW::getLog()->error("DNS host add error: $ip");
            $codes = IWorxError::extractErrorCodes($ok);

            $code = ($codes[0] > 0) ? $codes[0] : API::ERROR_DNSSYNC_ADDITION;
            return API::formatResponse($code);
        }

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Delete the given dns host (listener or publisher) from this box.
     *
     * @param array|struct $params
     * @return array|struct
     */
    public function dns_del_host($params)
    {
        API::auth($params);

        $ip1      = $params['ip'];
        $role     = $params['role'];
        $ip2      = $_SERVER['REMOTE_ADDR'];
        $hostname = gethostbyaddr($ip2);
        $force    = $params['force'];

        $q_ip1      = q($ip1);
        $q_role     = q($role);
        $q_ip2      = q($ip2);
        $q_hostname = q($hostname);

        $where = "(role = $q_role AND hostname = $q_ip1) OR
              (role = $q_role AND ipaddr = $q_ip1)   OR
              (role = $q_role AND hostname = $q_hostname) OR
              (role = $q_role AND ipaddr = $q_ip2)";

        IW::getLog()->debug("removing DNS host: $where");

        $host = new Dns_SyncHost();
        if (!$host->exists($where)) {
            IW::getLog()->debug("no such DNS host to delete: $where");
            return API::formatResponse(API::ERROR_DNSSYNC_NO_SUCH_HOST);
        }

        $host->get($where);

        $shm = new Dns_SyncHostManager();

        // WE DON'T WANT TO CASCADE HERE SINCE THE CASCADING
        // ONLY TAKES PLACE OUTSIDE OF THE API.  ONCE WE'RE
        // IN THE API WE NEED TO NOT CASCADE AS THIS WOULD
        // CAUSE AN INFINITE LOOP.

        if (($ok = $shm->deleteHost(
            $host,
            $force,
            $cascade = false
        ) !== true ) ) {
            IW::getLog()->error("DNS host delete ERROR: $where");
            $codes = IWorxError::extractErrorCodes($ok);

            $code = ($codes[0] > 0) ? $codes[0] : API::ERROR_DNSSYNC_DELETION;
            return API::formatResponse($code);
        }

        IW::getLog()->debug("DNS host delete SUCCESS: $where");

        return API::formatResponse(API::SUCCESS);
    }

    /**
     * Grab the dns data, returned in base64 encoded format from this box.
     *
     * @param array|struct $params
     * @return string
     */
    public function dns_get_data($params)
    {
        API::auth($params);

        $ip       = $_SERVER['REMOTE_ADDR'];
        $hostname = gethostbyaddr($ip);

        // see if we know about this listener
        $q_ip       = q($ip);
        $q_hostname = q($hostname);
        $where      = "(role = 'slave' AND ipaddr = $q_ip) OR
                   (role = 'slave' AND hostname = $q_hostname)";

        $host = new Dns_SyncHost();
        if ($host->exists($where)) {
            $host->get($where);

            // I DON'T WANT TO USE THE _NOW HERE AS WE'LL GET
            // TICKETS ABOUT WHY THE MASTER AND SLAVE HAVE SO
            // DIFFERENT STAMPS.. LET SEE IF MY GAMBLE PAYS OFF.

            $host->setLastSync(time());
        } else {
            $host = false;
        }

        IW::getLog()->debug("$ip requesting DNS sync");

        $data = Ini::get(Ini::DNS_DIR, 'datadir') . '/data-local';
        $copy = Ini::get(Ini::DNS_DIR, 'datadir') . '/data-sync-' . posix_getpid();

        if (!IWorxFileSys::isFile($data)) {
            return API::formatResponse(API::ERROR_DNSSYNC_NO_DATAFILE);
        }

        // make a backup of the sync file so we can zip it
        $ret = IWorxFileSys::cp(
            $data,
            $copy,
            $recurse = false,
            $force = true,
            $preserve = true
        );
        if ($ret != 0) {
            IW::getLog()->debug("$ip DNS sync FAILED: copy failed");
            if ($host !== false) {
                $host->setLastError('##LG_DNSSYNC_DATAFILE_COPY_FAILED##');
                $host->update();
            }
            return API::formatResponse(API::ERROR_DNSSYNC_EXPORT_ERROR);
        }

        // zip it
        $cmd = Ini::get(Ini::BIN, 'gzip') .
        " -f -9 {$copy} ";
        IWorxExec::exec($cmd, $result, $retval);

        if ($retval != 0) {
            IW::getLog()->debug("{$ip} DNS sync FAILED: gzip failed");
            if ($host !== false) {
                $host->setLastError('##LG_DNSSYNC_DATAFILE_COMPRESS_FAILED##');
                $host->update();
            }
            return API::formatResponse(API::ERROR_DNSSYNC_EXPORT_ERROR);
        }

        // grab it in a sendable format
        $zcopy = "{$copy}.gz";
        $file  = base64_encode(file_get_contents($zcopy));

        // log the successful request
        if ($host !== false) {
            IW::getLog()->debug("{$ip} DNS sync, updating last_sync on publisher");
            $host->setLastError('##LG_SUCCESS##');
            $host->update();
        } else {
            IW::getLog()->debug("{$ip} DNS sync, no such listener: [$where], " .
                     'not logging last_syn');
        }

        IW::getLog()->debug("{$ip} DNS sync, SUCCESS");
        @unlink($zcopy);
        return $file;
    }
}
