<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Jan 10, 2012
 */

/**
 * Form_SW_Cloudflare_Settings class.
 *
 * @package    InterWorx
 * @subpackage Form
 *
 */
class Form_SW_Cloudflare_Settings extends Form
{

    private $_stats;

    /**
     * GetInputForm function.
     */
    public function initInputForm()
    {
        $Form = new Form_Input();
        $Form->addInput(new Input_String('domain'))
             ->addFilter(new Zend_Filter_StringToLower());
        $this->setInputForm($Form);
    }

    /**
     * Initialize.
     */
    protected function _initialize()
    {
        parent::_initialize('cloudflare_settings_form');
        $this->setLabel('Cloudflare Settings');

        $plugin = IW::PluginManager()->getPlugin('cloudflare');
        if (IW::SW()->isParentReseller()) {
            $host_key = $plugin->getDatastoreVar('reseller_host_key');
        } else {
            $host_key = $plugin->getDatastoreVar('host_key');
        }
        $CF = new IWorx_CloudflareAPI($host_key);
        $user_data = $plugin->getDatastoreArray('user_key', IW::SW()->getId());
        list( $email, $user_key, $user_api_key ) = explode('|', $user_data);
        $domain = $this->getValue('domain');

        $API = new IWorx_CloudflareClientAPI($user_api_key);
        $stats = $API->stats($domain, $email);
        $this->_stats = $stats;

        $this->addInput($this->getInput('domain')->getDisabledCopy());

        $qh_security = 'CloudFlare provides security for your website and you can ' .
        'adjust your security setting for each website. A <b>low</b> security ' .
        'setting will challenge only the most threatening visitors. A <b>high</b> ' .
        'security setting will challenge all visitors that have exhibited ' .
        'threatening behavior within the last 14 days. We recommend starting with ' .
        'a high or medium setting.';

        $sec_lvl = $this->_getSecurityLevel($stats);
        $this->addInput(new Input_Select('security_level', $sec_lvl))
        ->setQuickHelp($qh_security)
        ->setLabel('Security Level')
        ->setDataSource(new DataSource_SecurityLevels());

        $qh_cache = 'Adjust your caching level to modify CloudFlare\'s caching ' .
        'behavior. The <b>basic</b> setting will cache most static resources ' .
        '(i.e., css, images, and JavaScript). The <b>aggressive</b> setting will ' .
        'cache all static resources, including ones with a query string.';

        $cache_lvl = $this->_getCacheLevel($stats);
        $this->addInput(new Input_Select('cache_level', $cache_lvl))
        ->setLabel('Cache Level')
        ->setQuickHelp($qh_cache)
        ->setDataSource(new DataSource_CacheLevels());

        $qh_dev = 'CloudFlare makes your website load faster by caching static ' .
        'resources like images, CSS and Javascript. If you are editing cachable ' .
        'content (like images, CSS, or JS) and want to see the changes right ' .
        'away, you should enter <b>Development mode</b>. This will bypass ' .
        'CloudFlare\'s cache. Development mode will automatically toggle off ' .
        'after <b>3 hours</b>. Hint: Press shift-reload if you do not see your ' .
        'changes immediate. If you forget to enter Development mode, you should ' .
        'log in to your CloudFlare.com account and use Cache Purge.';

        $dstype = DataSource_Boolean::ON_OFF;
        $dev_mode = ( $stats['dev_mode'] > 0 ) ? 1 : 0;
        $this->addInput(new Input_Set('dev_mode', $dev_mode))
        ->setHorizontalDisplay(true)
        ->setLabel('Dev Mode')
        ->setQuickHelp($qh_dev)
        ->setDataSource(DataSource_Boolean::factory($dstype));

        $qh_ipv6 = 'Automatically enable IPv6 networking for all your ' .
        'orange-clouded websites. CloudFlare will listen to IPv6 even if your ' .
        'host or server only supports IPv4.';

        $dstype = DataSource_Boolean::ENABLED_DISABLED;
        $toggle_ipv6 = ( $stats['toggle_ipv6'] > 0 ) ? 1 : 0;
        $this->addInput(new Input_Set('ipv6_enabled', $toggle_ipv6))
        ->setLabel('IPv6 Support')
        ->setQuickHelp($qh_ipv6)
        ->setHorizontalDisplay(true)
        ->setDataSource(DataSource_Boolean::factory($dstype));
    }

    /**
     * Form level validation.
     *
     * @return boolean
     */
    protected function _validate()
    {
        $error = ( $this->_stats['result'] === 'error' );
        if ($error) {
            $msg = 'CloudFlare is already hosting this domain under a different account.' .
            ' If you are the new, rightful owner, please contact CloudFlare and ' .
            'reference this message (#208).';
            $this->addError($msg);
            return false;
        }

        return true;
    }

    /**
     * Get sec_lvl.
     *
     * @param array $stats
     * @return string
     */
    private function _getSecurityLevel(array $stats)
    {
        $val = $stats['sec_lvl'];
        $levels = array( 'Off'    => 'eoff',
                     'High'   => 'high',
                     'Medium' => 'med',
                     'Low'    => 'low' );
        return $levels[ $val ];
    }

    /**
     * Get cache_lvl.
     *
     * @param array $stats
     * @return string
     */
    private function _getCacheLevel(array $stats)
    {
        $val = $stats['cache_lvl'];
        $levels = array( 'basic' => 'basic',
                     'agg'   => 'agg' );
        return $levels[ $val ];
    }
}
