<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Dec 16, 2011
 */

/**
 * CloudFlare API class.
 *
 * @package    InterWorx
 * @subpackage IWorx
 *
 */
class IWorx_CloudflareAPI
{
    // Debug settings
    const DEBUG_LEVEL    = 0; // 0 for no debug, 1 for debug, 2 for more debug, etc.
    const PRINT_RESPONSE = 1; // 1 to print the received response

    // Service URL
    const HOST_GW_SERVICE_URL = 'https://api.cloudflare.com/host-gw.html';

    // Clobber any previously assigned unique_ids. A 'unique_id' is a custom value,
    // an alias to map a user in your database to one in the CloudFlare system.
    // Any operations that can set a unique_id can be set to automatically "clobber"
    // or unassign a previously set unique_id.
    const ENABLE_CLOBBER_UNIQUE_ID = true;

    // For API access, specify your host_key:
    private $_hostKey;

    //cURL instance
    private $_cURL;

    /**
     * Constructor.
     *
     * @param string $hostkey
     */
    public function __construct($hostkey)
    {
        $this->_hostKey = $hostkey;
        $this->_initCurl();
    }

    /**
     * Create a CloudFlare account for a user. Register you as the host provider.
     *
     * @param string $email
     * @param string $pass
     * @param string $username
     * @param string $unique_id
     * @return array
     */
    public function userCreate($email, $pass, $username = null, $unique_id = null)
    {
        $params = [];
        $params['act'] = 'user_create';
        $params['cloudflare_email'] = $email;
        $params['cloudflare_pass'] = $pass;
        $params['cloudflare_username'] = $username;

        if ($unique_id !== null) {
            $params['unique_id'] = $unique_id;
        }

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Look a user by 'cloudflare_email' or by a previosly assigned 'unique_id'.
     *
     * @param string $type
     * @param string $value
     * @return array
     */
    public function userLookup($type, $value)
    {
        $params = [];
        $params['act'] = 'user_lookup';
        if ($type == 'cloudflare_email' || $type == 'unique_id') {
            $params[$type] = $value;
        } else {
            //IW::getLog()->error("ERROR: Invalid USER_LOOKUP type: '{$lookup_type}';" .
            //" Options are: cloudflare_email or unique_id\n" );
            return false;
        }

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Authorize you as the host provider for an existing user.
     *
     * @param string $email
     * @param string $pass
     * @param string $unique_id
     * @return array
     */
    public function userAuth($email, $pass, $unique_id = null)
    {
        $params = [];
        $params['act'] = 'user_auth';
        $params['cloudflare_email'] = $email;
        $params['cloudflare_pass']  = $pass;

        if ($unique_id !== null) {
            $params['unique_id'] = $unique_id;
        }

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Setup a Zone.
     *
     * @param string $user_key
     * @param string $zone_name
     * @param string $resolve_to
     * @param string $subdomains
     * @return array
     */
    public function zoneSet($user_key, $zone_name, $resolve_to, $subdomains)
    {
        $params = [];
        $params['act'] = 'zone_set';
        $params['user_key']   = $user_key;
        $params['zone_name']  = $zone_name;
        $params['resolve_to'] = $resolve_to;
        $params['subdomains'] = $subdomains;

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Lookup a Zone by 'user_key' and 'zone_name'.
     *
     * @param string $user_key
     * @param string $zone_name
     * @return array
     */
    public function zoneLookup($user_key, $zone_name)
    {
        $params = [];
        $params['act'] = 'zone_lookup';
        $params['user_key']  = $user_key;
        $params['zone_name'] = $zone_name;

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Delete a Zone by 'user_key' and 'zone_name'.
     *
     * @param string $user_key
     * @param string $zone_name
     * @return array
     */
    public function zoneDelete($user_key, $zone_name)
    {
        $params = [];
        $params['act'] = 'zone_delete';
        $params['user_key']  = $user_key;
        $params['zone_name'] = $zone_name;

        $ret = $this->_performRequest($params);
        return $ret;
    }

    /**
     * Initialize cURL variables.
     */
    private function _initCurl()
    {
        $ch = curl_init();

        curl_setopt($ch, CURLOPT_URL, self::HOST_GW_SERVICE_URL);

        if (self::DEBUG_LEVEL > 1) {
            curl_setopt($ch, CURLOPT_VERBOSE, 1);
        }

        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_TIMEOUT, 20);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_AUTOREFERER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

        $this->_cURL = $ch;
    }

    /**
     * Contact the service. Returns FALSE on error.
     *
     * @param array $data
     * @return array
     */
    private function _performRequest(array $data)
    {
        $data['host_key'] = $this->_hostKey;
        //IW::log( "CLIENT API REQUEST '{$data['act']}' INITIATED" );

        if (self::DEBUG_LEVEL > 0) {
            IW::getLog()->info('REQUEST DATA (to be sent via POST):');
            foreach ($data as $key => $value) {
                IW::getLog()->info("{$key}: {$value}");
            }
        }

        $ch = $this->_cURL;
        if ($data) {
            curl_setopt($ch, CURLOPT_POST, 1);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
        }

        if (( $http_result = curl_exec($ch) ) === false) {
            $error_string = 'A connectivity error occured while contacting the service.';
            $ret = array( 'msg' => $error_string );
            return $ret;
        }

        //curl_close( $ch );
        $ret = $this->_processResponse($http_result);
        $error = [];
        if ($ret['result'] === 'error') {
            $error['error'] = strip_tags($ret['msg']);
        } elseif (isset($ret['error'])) {
            $error['error'] = strip_tags($ret['error']);
        }
        //IW::log( "CLIENT API REQUEST '{$data['act']}' COMPLETED" );
        if (isset($error['error'])) {
            $error = array( 'error' => strip_tags($ret['msg']) );
            return $error;
        } else {
            return $ret['response'];
        }
    }

    /**
     * Process request results.
     *
     * @param mixed $response
     * @return array
     */
    private function _processResponse($response)
    {
        $decoded_response = json_decode($response, true);

        if (self::PRINT_RESPONSE) {
            $var_export = 'var_export';
            IW::getLog()->info($var_export($decoded_response, 1));
        }

        return $decoded_response;
    }
}
