<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Jan 9, 2012
 */

/**
 * CloudFlare Client API class.
 *
 * @package    InterWorx
 * @subpackage IWorx
 *
 */
class IWorx_CloudflareClientAPI
{
    // For API access, specify your host_key:
    private $_userAPIKey;

    //cURL instance
    private $_cURL;

    /**
     * Constructor.
     *
     * @param string $userAPIKey
     */
    public function __construct($userAPIKey)
    {
        $this->_userAPIKey = $userAPIKey;
    }

    /**
     * Retrieve the current stats and settings for a particular website.
     *
     * @param string $domain
     * @param string $email
     * @return array
     */
    public function stats($domain, $email)
    {
        $data = array(
                  'a' => 'stats',
                  'z' => $domain,
                  'email' => $email,
                  'interval' => 30
                  );

        $result = $this->_performRequest($data);
        $result = $this->_parseStatsResult($result);
        return $result;
    }

    /**
     * Parse stats results.
     *
     * @param array $result
     * @return array
     */
    private function _parseStatsResult(array $result)
    {
        $data = [];

        if ($result['result'] !== 'success') {
            $data['result'] = $result['result'];
            $data['msg']    = $result['msg'];
            $data['error']  = $result['msg'];
            return $data;
        }

        $traffic = $result['response']['result']['objs'][0]['trafficBreakdown'];
        $data['views_regular']   = $traffic['pageviews']['regular'];
        $data['views_threat']    = $traffic['pageviews']['threat'];
        $data['views_crawler']   = $traffic['pageviews']['crawler'];
        $data['uniques_regular'] = $traffic['uniques']['regular'];
        $data['uniques_threat']  = $traffic['uniques']['threat'];
        $data['uniques_crawler'] = $traffic['uniques']['crawler'];

        $bandwidth = $result['response']['result']['objs'][0]['bandwidthServed'];
        $data['bandwidth_cloudflare'] = $bandwidth['cloudflare'] * 1000;
        $data['bandwidth_user']       = $bandwidth['user'] * 1000;

        $requests = $result['response']['result']['objs'][0]['requestsServed'];
        $data['requests_cloudflare'] = $requests['cloudflare'];
        $data['requests_user']       = $requests['user'];

        $basic = $result['response']['result']['objs'][0];
        $data['sec_lvl']     = $basic['userSecuritySetting'];
        $data['cache_lvl']   = $basic['cache_lvl'];
        $data['dev_mode']    = ( $basic['dev_mode'] > 0 ) ? 1 : 0;
        $data['toggle_ipv6'] = $basic['ipv46'];

        $data['result'] = $result['result'];
        $data['msg']    = $result['msg'];

        return $data;
    }

    /**
     * Sets the Basic Security Level to HIGH / MEDIUM / LOW / ESSENTIALLY OFF.
     *
     * @param string $domain
     * @param string $email
     * @param string $sec_lvl
     * @return array
     */
    public function setSecurityLevel($domain, $email, $sec_lvl)
    {
        // v must be one of low|med|high|eoff

        $data = array(
                  'a' => 'sec_lvl',
                  'z' => $domain,
                  'email' => $email,
                  'v' => $sec_lvl
                  );

        $result = $this->_performRequest($data);
        return array( 'result' => $result['result'],
                  'msg'    => $result['msg'] );
    }

    /**
     * Sets the Caching Level to Aggressive or Basic.
     *
     * @param string $domain
     * @param string $email
     * @param string $cache_lvl
     * @return array
     */
    public function setCacheLevel($domain, $email, $cache_lvl)
    {
        // v must be one of agg|basic

        $data = array(
                  'a' => 'cache_lvl',
                  'z' => $domain,
                  'email' => $email,
                  'v' => $cache_lvl
                  );

        $result = $this->_performRequest($data);
        return array( 'result' => $result['result'],
                  'msg'    => $result['msg'] );
    }

    /**
     * Toggle Development Mode on or off for a particular domain.
     *
     * @param string $domain
     * @param string $email
     * @param string $status
     * @return array
     */
    public function toggleDevMode($domain, $email, $status)
    {
        $data = array(
                  'a' => 'devmode',
                  'z' => $domain,
                  'email' => $email,
                  'v' => $status
                  );

        $result = $this->_performRequest($data);
        return array( 'result' => $result['result'],
                  'msg'    => $result['msg'] );
    }

    /**
     * Purge CloudFlare of any cached files.
     *
     * @param string $domain
     * @param string $email
     * @return array
     */
    public function purgeCache($domain, $email)
    {
        $data = array(
                  'a' => 'fpurge_ts',
                  'z' => $domain,
                  'email' => $email,
                  'v' => '1'
                  );

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Check whether one or more websites/domains are active under an account.
     *
     * @param array  $zones
     * @param string $email
     * @return array
     */
    public function zoneCheck(array $zones, $email)
    {
        $domains = implode(',', $zones);
        $data = array(
                  'a' => 'zone_check',
                  'zones' => $domains,
                  'email' => $email
                  );

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Returns a list of IP addresses which hit your site classified by type.
     *
     * @param string  $zone_id
     * @param string  $email
     * @param string  $hours
     * @param string  $class
     * @param boolean $geo
     * @return array
     */
    public function zoneIPs($zone_id, $email, $hours, $class, $geo = null)
    {
        // class must be r|s|t
        if ($hours === null) {
            $hours = '24';
        }

        $data = array(
                  'a' => 'zone_ips',
                  'zid' => $zone_id,
                  'email' => $email,
                  'hours' => $hours,
                  '$class' => $class
                  );

        if ($geo !== null) {
            $data['geo'] = '1';
        }

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Tells CloudFlare to take a new image of your site.
     *
     * @param string $zone_id
     * @param string $email
     * @return array
     */
    public function updateSnapshot($zone_id, $email)
    {
        $data = array(
                  'a' => 'zone_grab',
                  'zid' => $zone_id,
                  'email' => $email
                  );

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Whitelist an IP address.
     *
     * @param string $domain
     * @param string $email
     * @return array
     */
    public function whitelistIP($domain, $email)
    {
        $data = array(
                  'a' => 'wl',
                  'key' => $domain,
                  'u' => $email
                  );

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Blacklist an IP address.
     *
     * @param string $domain
     * @param string $email
     * @return array
     */
    public function blacklistIP($domain, $email)
    {
        $data = array(
                  'a' => 'ban',
                  'key' => $domain,
                  'u' => $email
                  );

        $result = $this->_performRequest($data);
        return $result;
    }

    /**
     * Toggles ipv6 support for a site.
     *
     * @param string $domain
     * @param string $email
     * @param string $status
     * @return array
     */
    public function toggleIPv6($domain, $email, $status)
    {
        // enable toggle was changed from 1 to 3
        // see ticket #3932
        $status = ( $status == 1 ) ? 3 : 0;
        $data   = array(
                    'a' => 'ipv46',
                    'z' => $domain,
                    'u' => $email,
                    'v' => $status
                    );

        $result = $this->_performRequest($data);
        return array( 'result' => $result['result'],
                  'msg'    => $result['msg'] );
    }

    /**
     * Initialize cURL environment.
     */
    private function _initCurl()
    {
        $url = 'https://www.cloudflare.com/api_json.html';
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_FORBID_REUSE, true);
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_POST, 1);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);

        $this->_cURL = $ch;
    }

    /**
     * Contact the service. Returns FALSE on error.
     *
     * @param array $data
     * @return array
     */
    private function _performRequest(array $data)
    {
        $this->_initCurl();
        $data['tkn'] = $this->_userAPIKey;
        $ch = $this->_cURL;

        curl_setopt($ch, CURLOPT_POSTFIELDS, $data);

        if (( $http_result = curl_exec($ch) ) === false) {
            $error_string = 'A connectivity error occured while contacting the service.';
            $ret = array( 'msg' => $error_string );
            return $ret;
        }

        $error = curl_error($ch);
        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        if ($http_code != 200) {
            $error = array( 'error' => strip_tags($error) );
            return $error;
        }
        $result = json_decode($http_result, true);

        return $result;
    }

    /**
     * Formatter for bandwidth units.
     *
     * @param integer $val
     * @return string
     */
    private function _getLabel($val)
    {
        return IWorxUtil::formatHumanReadableSize($val, null, false, '%01.0f');
    }
}
