<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Dec 14, 2011
 */

/**
 * CloudFlare integration plugin.
 *
 * @package    InterWorx
 * @subpackage Plugin
 */
class Plugin_Cloudflare extends Plugin
{
    /**
     * Small menu icon - uses URL path, not system path.
     *
     * @var string
     */
    protected $_sw_menu_icon_small = '/plugins/images/cloudflare/cloudflare.gif';

    /**
     * Large menu icon - uses URL path, not system path.
     *
     * @var string
     */
    protected $_sw_menu_icon_large = '/plugins/images/cloudflare/cloudflare-big.gif';

    /**
     * Init my edit form.
     *
     * Add plugin settings.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function initializeMyEditForm(Form_NW_Plugins_Edit $Form)
    {

        $Group = $Form->addGroup('extra_options', 'Extra Plugin Settings');

        $host_key = $this->getDatastoreVar('host_key');
        $Group->addInput(new Input_String('host_key', $host_key))
        ->setRequired(true)
        ->setQuickhelp('Contact Cloudflare at http://www.cloudflare.com/ to receive' .
                      ' a host key.');

        $reseller_host_key = $this->getDatastoreVar('reseller_host_key');
        $Group->addInput(new Input_String('reseller_host_key', $reseller_host_key))
        ->setRequired(false)
        ->setQuickHelp('A reseller host key is a generically branded host key that ' .
                      'is still bound to your account with cloudflare, but ' .
                      'communication from cloudflare to the client will reference ' .
                      '"Your Provider" as opposed to "Company X".  Contact ' .
                      'Cloudflare at http://www.cloudflare.com/ to receive a '.
                      'reseller host key, if desired.');

        $allow_resellers = $this->getDatastoreVar('allow_resellers');
        $Group->addInput(new Input_Flag('allow_resellers', $allow_resellers))
        ->setLabel('##LG_ALLOW_RESELLER_ACCESS_TO_CLOUDFLARE##');
    }

    /**
     * Process my edit form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function processMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $host_key          = $Form->getValue('host_key');
        $allow_resellers   = $Form->getValue('allow_resellers');
        $reseller_host_key = $Form->getValue('reseller_host_key');

        $this->setDatastoreVar('host_key', $host_key);
        $this->setDatastoreVar('allow_resellers', $allow_resellers);
        $this->setDatastoreVar('reseller_host_key', $reseller_host_key);
    }

    /**
     * Customizations to siteworx menu can be done here.
     *
     * @param IWorxMenuManager $MenuMan
     */
    public function updateSiteworxMenu(IWorxMenuManager $MenuMan)
    {
        try {
            if (IW::SW()->getWorkingUser()->getMenuStyle() === User::MENU_BIG) {
                $icon = $this->_sw_menu_icon_large;
            } else {
                $icon = $this->_sw_menu_icon_small;
            }
        } catch (IWorx_Exception_NotLoggedIn $e) {
            $icon = $this->_sw_menu_icon_small;
        }

        $new_data = array( 'text'   => 'CloudFlare',
                       'url'    => '/siteworx/cloudflare',
                       'parent' => 'iw-menu-features-domains',
                       'img'  => $icon );

        $allow_resellers = $this->getDatastoreVar('allow_resellers');
        if (!IW::SW()->isParentReseller() || $allow_resellers) {
            $MenuMan->addMenuItemAfter('iw-menu-features-dns', 'cloudflare', $new_data);
        }
    }

    /**
     * Returns true if the cloudflare users is already setup for this siteworx acct.
     *
     * @return boolean
     */
    public function isCloudflareUserSetup()
    {
        $data = $this->getDatastoreArray('user_key', IW::SW()->getId());
        return $data !== null;
    }

    /**
     * Returns the cloudflare user e-mail (assuming it is already setup for this acct).
     *
     * @return string|boolean
     */
    public function getCloudflareUserEmail()
    {
        $data = $this->getDatastoreArray('user_key', IW::SW()->getId());
        if (isset($data[0])) {
            list( $email, $key ) = explode('|', $data);
            $result = $email;
        } else {
            $result = false;
        }
        return $result;
    }

    /**
     * Returns the cloudflare user-key data.
     *
     * @return array
     */
    public function getUserKeyData()
    {
        return explode('|', $this->getDatastoreArray('user_key', IW::SW()->getId()));
    }

    /**
     * Sets the cloudflare user-key data.
     *
     * @param array $data
     */
    public function setUserKeyData(array $data)
    {
        $this->setDatastoreArray(
            'user_key',
            IW::SW()->getId(),
            implode('|', $data)
        );
    }

    /**
     * Unsets the cloudflare user-key data.
     */
    public function clearUserKeyData()
    {
        $this->unsetDatastoreArray('user_key', IW::SW()->getId());
    }

    /**
     * Sets the cloudflare domain status data.
     *
     * @param string $subdomain
     * @param string $status
     */
    public function setDomainStatusData($subdomain, $status)
    {
        $this->setDatastoreArray('status', $subdomain, $status);
    }

    /**
     * Get the host API object, used for calling cloudflare host api.
     *
     * @return IWorx_CloudflareAPI
     */
    public function getHostAPI()
    {
        if (IW::SW()->isParentReseller()) {
            $host_key = $this->getDatastoreVar('reseller_host_key');
        } else {
            $host_key = $this->getDatastoreVar('host_key');
        }
        return new IWorx_CloudflareAPI($host_key);
    }

    /**
     * Get category.
     *
     * @return string
     */
    public function getCategory()
    {
        return Plugin_Category::CDN;
    }
}
