<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Paul Oehler <poehler@interworx.com>
 * @date    Sep 27, 2009
 */

/**
 * Plugin for execution of 3rd party scripts after a successful action completes.
 *
 * @package    InterWorx
 * @subpackage Plugin
 */
class Plugin_Eventhooks extends Plugin
{
    private $_hook_config = '/etc/iworx-hooks.conf';

    /**
     * Plugin initiation.
     */
    protected function _init()
    {
        $hook_config = $this->getDatastoreVar('hook_config');

        if ($hook_config !== '') {
            $this->_hook_config = $hook_config;
        }
    }

    /**
     * Get priority.
     *
     * @return integer
     */
    public function getPriority()
    {
        return 40;
    }

    /**
     * Get the plugin category.
     *
     * @return string
     */
    public function getCategory()
    {
        return Plugin_Category::ADVANCED;
    }

    /**
     * Initialize this plugin's form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function initializeMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $Group = $Form->addGroup('extra_options', 'Extra Plugin Settings');

        $qh = 'Download a sample hooks file here: ' .
        'http://interworx.com/iworx-hooks.conf';
        $Group->addInput(new Input_String('hook_config', $this->_hook_config))
        ->setLabel('Hook Configuration File')
        ->setHint('File must be readable by "iworx" user')
        ->setDescription('File on server containing hook configuration');

        $Group->addInput(new Input_String('samplefile', $qh, false))
        ->setLabel('Sample File');
    }

    /**
     * Process the form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function processMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $this->setDatastoreVar('hook_config', $Form->getValue('hook_config'));
    }

    /**
     * This method is called after all controller actions.
     *
     * @param string        $ctrl_act
     * @param Ctrl_Abstract $Ctrl
     * @param string        $action
     * @param mixed         $params
     */
    public function postAction($ctrl_act, Ctrl_Abstract $Ctrl, $action, $params)
    {
        putenv(sprintf("iw_controller=%s", get_class($Ctrl)));

        // if it's not a Commit action, we don't care about it.
        if (substr($action, -6) !== 'Commit') {
            return;
        }

        // if the action wasn't successful, we don't care about it
        if (!$Ctrl->isViewSuccess()) {
            return;
        }

        // verify readability
        $hc = $this->_hook_config;
        if (is_readable($hc) === false) {
            IW::getLog()->info('hook config not readable');
            return;
        }

        // start processing the config file
        $hc_file = file($hc);
        foreach ($hc_file as $line) {
            $result = [];
            $retval = 0;
            $line   = trim($line);

            // skip comments
            if (substr($line, 0, 1) === '#') {
                continue;
            }

            // If this controller/action combo doesn't match the current line, skip it
            list( $ctrl, $act, $prog_and_args ) = preg_split('/[\s]+/', $line, 3);
            if ($ctrl_act !== "{$ctrl}:{$act}Commit") {
                continue;
            }

            // make sure the specified script is executable by iworx, if not bail
            list( $prog, $args ) = preg_split('/[\s]+/', $prog_and_args, 2);
            if (is_executable($prog) === false) {
                IW::getLog()->info('hook prog not executable by iworx: ' . $prog);
                continue;
            }

            // prepare environment variables
            if ($params instanceof Form) {
                $Form = $params;
                $this->_addInputsToEnv($Form->getInputForm());
                $this->_addInputsToEnv($Form);
            }

            putenv("iw_action={$action}");

            if ($Ctrl instanceof Ctrl_Siteworx_Abstract) {
                try {
                    putenv('iw_working_domain=' . IW::SW()->getWorkingDomainName());
                    putenv('iw_master_email=' . IW::SW()->getMasterUser()->getEmail());
                } catch (IWorx_Exception $e) {
                    IW::getLog()->error('eventhooks: failed to retrieve working domain!');
                }
            }

            // go! go! go!
            exec($prog_and_args, $result, $retval);

            if ($retval !== 0) {
                $output = implode("\n", $result);
                IW::getLog()->info('plugin hook exited non-zero: ' . $prog_and_args . ': ' . $output);
            }
        }//end foreach line in the iworx-hooks.conf file
    }

    /**
     * Add inputs to os environment.
     *
     * @param Form $Form
     */
    private function _addInputsToEnv(Form $Form)
    {
        foreach ($Form as $Group) {
            foreach ($Group as $Input) {
                /* @var $Input Input_Input */
                foreach ($Input->getAliases() as $alias) {
                    $value = $Input->getCurrentValue();
                    if (is_array($value)) {
                        $value = implode(',', $value);
                    }
                    putenv("iw_{$alias}={$value}");
                }//end foreach
            }//end foreach
        }//end foreach
    }//end function
}
