<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Feb 27, 2012
 */

/**
 * External NFS Quota management plugin.
 *
 * @package    InterWorx
 * @subpackage Plugin
 */
class Plugin_ExternalNfsQuota extends Plugin
{
    private $_last_error;
    private $_last_retval;

    /**
     * Return the plugin category.
     *
     * @return string
     */
    public function getCategory()
    {
        return Plugin_Category::ADVANCED;
    }

    /**
     * Initialize this plugin's form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function initializeMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $hostname = $this->getDatastoreVar('nfs_host');
        $homedir = $this->getDatastoreVar('nfs_homedir');

        $Group = $Form->addGroup('extra_options', 'Extra Plugin Settings');
        $Group->addInput(new Input_String('hostname', $hostname))
        ->setLabel('NFS Hostname or IP')
        ->setQuickHelp('##LG_QH_NFS_HOSTNAME##')
        ->setRequired(false);
        $Group->addInput(new Input_String('home_directory', $homedir))
        ->setLabel('NFS Quota Partition')
        ->setQuickHelp('##LG_QH_NFS_PARTITION##')
        ->setRequired(false);
        $Group->addInput(new Input_Password())
        ->setLabel('NFS Root Password')
        ->setQuickHelp('##LG_QH_NFS_PASSWORD##')
        ->setRequired(false);
    }

    /**
     * Process the form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function processMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $hostname = $Form->getValue('hostname');
        $homedir  = $Form->getValue('home_directory');
        $password = $Form->getValue('password');

        $this->setDatastoreVar('nfs_host', $hostname);
        $this->setDatastoreVar('nfs_homedir', $homedir);
        if ($Form->getValue('password') !== '') {
            $password = IWorxEncrypt::encrypt($password);
            $this->setDatastoreVar('nfs_password', $password);
        }
    }

    /**
     * Form validation.
     *
     * @param Form $Form
     * @return boolean
     */
    public function validateMyEditForm(Form $Form)
    {
        if ($Form->getValue('status') === '0') {
            return true;
        }

        $hostname = $Form->getValue('hostname');
        $homedir  = $Form->getValue('home_directory');
        $password = $Form->getValue('password');

        if ($password === '') {
            $password = $this->getDatastoreVar('nfs_password');
            if ($password !== '') {
                $password = IWorxEncrypt::decrypt($password);
            }
        }

        $SSH = new IWorxSSH($hostname, 'root', $password);
        $connect = $SSH->testConnection();

        if ($connect !== 'success') {
            $Form->addError('Unable to connect to remote server.');
            return false;
        }

        $cmd = "test -e {$homedir} && echo 'success'";
        if ($SSH->sshCommand($cmd) !== 'success') {
            $Form->addError('Specified directory does not exist.');
            return false;
        }

        $cmd = "test -d {$homedir} && echo 'success'";
        if ($SSH->sshCommand($cmd) !== 'success') {
            $Form->addError('Specified directory does not exist.');
            return false;
        }

        return true;
    }

    /**
     * Check to see if the connection is configured properly.
     *
     * @return boolean $connected
     */
    public function testConnection()
    {
        $hostname = $this->getDatastoreVar('nfs_host');
        $password = $this->getDatastoreVar('nfs_password');
        $password = IWorxEncrypt::decrypt($password);

        $SSH       = new IWorxSSH($hostname, 'root', $password);
        $connected = ( $SSH->testConnection() === 'success' );

        return $connected;
    }

    /**
     * Execute a command on the NFS.
     *
     * @param string  $cmd     command to be run
     * @param array   $result output
     * @param integer $retval exit status
     */
    private function _sshCommand($cmd, array &$result = null, &$retval = null)
    {
        $host     = $this->getDatastoreVar('nfs_host');
        $password = $this->getDatastoreVar('nfs_password');
        $password = IWorxEncrypt::decrypt($password);
        $SSH      = new IWorxSSH($host, 'root', $password);

        $result = explode("\n", $SSH->sshCommand($cmd));
        $retval = ( $SSH->isError() ) ? 1 : 0;
        if ($retval === 1) {
            IW::getLog()->warning($result);
        }
    }

    /**
     * Add a unix group.
     *
     * @param string  $ACT_GID
     * @param string  $uniqname
     * @param integer $retval
     */
    public function groupAdd($ACT_GID, $uniqname, &$retval = null)
    {
        $cmd = Ini::get(Ini::BIN, 'groupadd') . " -g {$ACT_GID} {$uniqname}";
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Add a unix user.
     *
     * @param string  $homedir
     * @param string  $ACT_UID
     * @param string  $ACT_GID
     * @param string  $uniqname
     * @param array   $result
     * @param integer $retval
     */
    public function userAdd(
        $homedir,
        $ACT_UID,
        $ACT_GID,
        $uniqname,
        array &$result = null,
        &$retval = null
    ) {

        $cmd = Ini::get(Ini::BIN, 'useradd') . " -d {$homedir} " . '-s /sbin/nologin ';

        if ($ACT_UID !== false) {
            $cmd .= " -u {$ACT_UID} ";
        }

        if ($ACT_GID !== false) {
            $cmd .= " -g {$ACT_GID} ";
        }

        $cmd .= " {$uniqname} ";

        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Delete a unix user.
     *
     * @param string  $uniqname
     * @param array   $result
     * @param integer $retval
     */
    public function userDel($uniqname, array &$result = null, &$retval = null)
    {
        $cmd = Ini::get(Ini::BIN, 'userdel') . " -f {$uniqname} ";
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Check a group quota.
     *
     * @param string  $uniqname
     * @param array   $result
     * @param integer $retval
     */
    public function quota($uniqname, array &$result = null, &$retval = null)
    {
        $cmd = Ini::get(Ini::BIN, 'quota') . " -g {$uniqname}";
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Get all group quotas.
     *
     * @param array   $result
     * @param integer $retval
     */
    public function repQuota(array &$result = null, &$retval = null)
    {
        $cmd = Ini::get(Ini::BIN, 'repquota') . ' -g ' . $this->getHomeDir();
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Set a group quota.
     *
     * @param string  $uniqname
     * @param integer $soft
     * @param integer $hard
     * @param array   $result
     * @param integer $retval
     */
    public function setQuota(
        $uniqname,
        $soft,
        $hard,
        array &$result = null,
        &$retval = null
    ) {
        $cmd = Ini::get(Ini::BIN, 'setquota') .
        " -g {$uniqname} {$soft} {$hard} 0 0 " . $this->getHomeDir();
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Set the default quota block grace time for quota partition.
     *
     * @param integer $seconds
     * @param array   $result
     * @param integer $retval
     */
    public function setQuotaGrace($seconds, array &$result = null, &$retval = null)
    {
        $cmd = Ini::get(Ini::BIN, 'setquota') .
        " -t -g {$seconds} {$seconds} {$this->getHomeDir()}";
        $this->_sshCommand($cmd, $result, $retval);
    }

    /**
     * Get the NFS mount directory.
     *
     * @return string
     */
    public function getHomeDir()
    {
        return $this->getDatastoreVar('nfs_homedir');
    }
}
