<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @subpackage Payload
 * @author     Jon Daniel <jdaniel@interworx.com>
 * @date       Feb 21, 2011
 */

/**
 * Displays the page uri based on the ctrl and action name.
 *
 * @package    InterWorx
 * @subpackage Payload
 */
class IWorx_Formatter_History_Change extends IWorx_Formatter_NoneIfEmpty
{

    /**
     * Format.
     *
     * @param mixed  $value The value being rendered
     * @param object $row   The data for the entire row
     * @param array  $data  The entire dataset of the Payload
     *
     * @return string
     */
    public function format($value, $row, array $data)
    {
        $from = $this->_getValue($row->initial_value);
        $to   = $this->_getValue($row->current_value);

        if (is_array($from) === true || is_array($to) === true) {
            $html = $this->_getArrayDiffString(
                $row->input_label,
                (array)$from,
                (array)$to
            );
        } elseif ($from != '' || $to != '') {
            $html = $this->_getFieldChangedString(
                $row->input_label,
                $from,
                $to
            );
        }

        if (empty($html)) {
            $html = $this->getNoneString();
        }

        return $html;
    }

    /**
     * Make the given text bold.
     *
     * @param string $value
     * @return string
     */
    private function _bold($value)
    {
        return "<b>{$value}</b>";
    }

    /**
     * Determines if the value is a JSON encoded array.
     *
     * And returns the correct value.
     *
     * @param string $value
     * @return string|array
     */
    private function _getValue($value)
    {
        $unlimited_val = Ini::get(Ini::CONSTANTS, 'unlimited');
        if ($this->_isJson($value) === true) {
            $value = json_decode($value);
        }

        return ( $value == $unlimited_val ) ? '##LG_UNLIMITED##' : $value;
    }

    /**
     * Gets the language string for the array diff.
     *
     * @param string $lang_key
     * @param array  $from_array
     * @param array  $to_array
     * @return string
     */
    private function _getArrayDiffString(
        $lang_key,
        array $from_array = null,
        array $to_array = null
    ) {

        $lang_key = $this->_bold(IW::tr($lang_key));
        $html     = [];

        if (count($from_array) > count($to_array)) {
            $diff_array = array_diff($from_array, $to_array);
            $array_str  = $this->_threeThenNewline($diff_array);

            $html[] = "##LG_FIELD_REMOVED_FROM|{$array_str}|{$lang_key}##";
        }

        if (count($from_array) < count($to_array)) {
            $diff_array = array_diff($to_array, $from_array);
            $array_str  = $this->_threeThenNewline($diff_array);

            $html[] = "##LG_FIELD_ADDED_TO|{$array_str}|{$lang_key}##";
        }

        return implode('<br>', $html);
    }

    /**
     * Three array elements then a new line.
     *
     * @param array $array
     * @return string
     */
    private function _threeThenNewline(array $array)
    {
        $output = [];

        foreach ($array as $key => $value) {
            $array[$key] = $this->_bold($value);
        }

        return implode(', ', $array);
    }

    /**
     * Gets the string that indicates how the field was changes.
     *
     * @param string $lang_key
     * @param string $from
     * @param string $to
     * @return string
     */
    private function _getFieldChangedString($lang_key, $from, $to)
    {
        $initial_from = $from;

        $lang_key = IW::tr($this->_bold($lang_key));

        if ($from !== '') {
            $from = IW::tr($this->_bold($from));
        } else {
            $from = IW::tr($this->getNoneString());
        }

        if ($to !== '') {
            $to = IW::tr($this->_bold($to));
        } else {
            $to = IW::tr($this->getNoneString());
        }

        $from_first_newline = strpos($from, "\n");
        $to_first_newline   = strpos($to, "\n");

        if ($from_first_newline !== false) {
            $from = substr($from, 0, $from_first_newline) . '...';
        }

        if ($to_first_newline !== false) {
            $to = substr($to, 0, $to_first_newline) . '...';
        }

        if (empty($initial_from)) {
            $html = "##LG_FIELD_SET_TO|{$lang_key}|{$to}##";
        } else {
            $html = "##LG_FIELD_CHANGED_FROM_TO|{$lang_key}|{$from}|{$to}##";
        }

        return $html;
    }

    /**
     * Checks to see if the specified string is a JSON array.
     *
     * @param string $value
     * @return boolean
     */
    private function _isJson($value)
    {
        $tmp = json_decode($value);
        return json_last_error() == 0;
    }
}
