<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Paul Oehler <poehler@interworx.com>
 * @date    Sep 22, 2009
 */

/**
 * Let's Encrypt integration plugin.
 *
 * @package    InterWorx
 * @subpackage Plugin
 */
class Plugin_LetsEncrypt extends Plugin
{
    private $_last_error = '';

    /**
     * Init the plugin, disabling if necessary.
     */
    protected function _init()
    {
        if ($this->_checkOS() === false) {
            $this->_disable($this->_last_error);
        }
    }

    /**
     * Initialize this plugin's form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function initializeMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $email = $this->getDatastoreVar('expire_email');
        $autossl = Ini::get(Ini::APACHE, 'autossl');
        $le_mode = Ini::get(Ini::APACHE, 'le_mode');

        $Group = $Form->addGroup('extra_options', 'Extra Plugin Settings');

        //if ini isn't set make it on by default
        if ($autossl === null) {
            $autossl = 1;
        }

        if ($le_mode === null) {
            $le_mode = 'live';
        }

        $Group->addInput(new Input_Boolean_YesNo('autossl', $autossl))
          ->setLabel('Autossl')
          ->setQuickHelp('Attempt to generate an SSL certificate for all SiteWorx account domains.');

        $DS = new DataSource_ArrayAssoc(['staging' => 'Staging',
                                         'live' => 'Live']);
        $Group->addInput(new Input_Select('le_mode', $le_mode))
          ->setDataSource($DS)
          ->setQuickHelp('##LG_QH_LETS_ENCRYPT_MODE##')
          ->setLabel('##LG_MODE##');

        $Group->addInput(new Input_Email('email', $email))
              ->setLabel('Expiration Warning Email')
              ->setQuickHelp('##LG_LETS_ENCRYPT_EXPIRATION_WARNING_EMAIL##')
              ->addValidator(new IWorx_Validate_Email());
    }

    /**
     * Install the plugin.
     */
    public function install()
    {
        IW::getLog()->info('Installing ' . get_class($this));
        $basedir     = Ini::get(Ini::IWORX_DIR, 'base');
        $setsid      = Ini::get(Ini::BIN, 'setsid');
        $timeout     = 300;
        $letsencrypt = '/home/interworx/bin/install-letsencrypt.sh';
        $timeout_cmd = "{$setsid} {$basedir}/bin/exec_timeout.sh -t {$timeout} ";
        $runuser_cmd = Ini::get(Ini::IWORX_BIN, 'runasuser') . ' root custom ';

        $final_cmd = "{$timeout_cmd} {$runuser_cmd} '{$letsencrypt}'";
        IW::getLog()->info("Running {$final_cmd}");

        IW::getLog()->info($final_cmd);

        IWorxExec::exec($final_cmd, $result, $retval, IWorxExec::STDERR_2_STDOUT);
        IW::getLog()->info($retval);
        IW::getLog()->info($result);
    }

    /**
     * Form validation.
     *
     * @param Form $Form
     * @return boolean
     */
    public function validateMyEditForm(Form $Form)
    {
        if ($Form->getValue('status') === '0') {
            return true;
        }
        if ($this->_checkOS() === false) {
            $Form->addError($this->_last_error);
        }
        return !$Form->hasErrors();
    }

    /**
     * Process the form.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function processMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $email = $Form->getValue('email');
        $autossl = $Form->getValue('autossl');
        $le_mode = $Form->getValue('le_mode');
        $status = $Form->getValue('status');
        if ($this->getDatastoreVar('expire_email') != $email) {
            $cmd = Ini::get(Ini::BIN, 'letsencrypt') .  ' register --noninteractive'
                   . ' --update-registration --email ' . $email;

            IWorxExec::exec($cmd, $result, $retval);
            $this->setDatastoreVar('expire_email', $email);
        }

        Ini::set(Ini::APACHE, 'autossl', $autossl);
        Ini::set(Ini::APACHE, 'le_mode', $le_mode);
        Ini::write();

        if ($status == false) {
            Ini::set(Ini::APACHE, 'autossl', 0);
            Ini::write();
        }
    }

    /**
     * Determine if the OS is supported by LetsEncrypt.
     *
     * Sets $this->_last_error to a string error message.
     *
     * @return boolean
     */
    private function _checkOS()
    {
        $distro = IWorxSysInfo::distro();
        $found  = preg_match('/release ([0-9]+)/', $distro, $matches);
        if ($found === 0) {
            $this->_last_error = 'Could not determine OS version';
        } else {
            $version = intval($matches[1]);
            if ($version <= 5) {
                $this->_last_error = "Let's Encrypt is not supported on your OS: {$distro}";
            }
        }

        return $this->_last_error === '';
    }

    /**
     * Get category.
     *
     * @return string
     */
    public function getCategory()
    {
        return Plugin_Category::SSL;
    }
}
