<?php
/**
 * InterWorx Hosting Control Panel
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2016 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @subpackage Upgrade
 * @author     Tim College <tcollege@interworx.com>
 * @date       Feb 24, 2016
 */

namespace Interworx;

use \Psr\Log\LoggerInterface;
use Interworx\CliCore\PathFilter;

/**
 * Class CliCore
 *
 * @package    interworx-cli-core
 * @subpackage CLI
 */
abstract class CliCore
{
    const PARAM_REQUIRES_CODE = 101;
    const NO_ACTION_FOUND     = 123;
    const OK                  = 0;
    const NOT_OK              = 1;
    const INVALID_INPUT       = 2;
    const NOT_AUTHORIZED      = 3;

    private $_functions = [];
    private $_opt_args  = [];
    private $_paths     = [];

    private $_ColorConsole;

    /**
     * Getopt object that handles command line parameter parsing.
     *
     * @var \Zend_Console_Getopt
     */
    static private $_opts = null;

    /**
     * PSR-3 compliant logger.
     *
     * @var \Psr\Log\LoggerInterface
     */
    private $_Logger;

    /**
     * CliCore constructor.
     *
     * @param \Psr\Log\LoggerInterface $Logger
     */
    public function __construct(LoggerInterface $Logger)
    {
        $this->setLogger($Logger);
        $this->_ColorConsole = new \Console_Color2();
        $this->_logScriptStart();
        $this->registerOptArgs();
    }

    /**
     * Log the beginning of the script.
     */
    protected function _logScriptStart()
    {
        $this->getLogger()->info('script begin');
        $this->getLogger()->notice(implode(' ', $this->_getArgv()));
    }

    /**
     * Gets the Logger object.
     *
     * @return \Psr\Log\LoggerInterface
     */
    public function getLogger()
    {
        return $this->_Logger;
    }

    /**
     * Sets the logger.
     *
     * @param \Psr\Log\LoggerInterface $Logger
     * @throws CliCore\Exception If logger has already been set.
     */
    public function setLogger(LoggerInterface $Logger)
    {
        if ($this->_Logger !== null) {
            throw new CliCore\Exception('Logger is already set, cannot override');
        }
        $this->_Logger = $Logger;
    }

    /**
     * Get the arguments passed into the script.
     *
     * @return array
     */
    protected function _getArgv()
    {
        global $argv;
        return (is_array($argv)) ? $argv : [];
    }

    /**
     * Register known functions and parameters.
     */
    public function registerOptArgs()
    {

    }

    /**
     * Turns off output buffering.
     */
    public function disableOutputBuffer()
    {
        // Many places show doing this twice to be safe.  Dunno.
        @ob_end_flush();
        @ob_end_flush();
    }

    /**
     * Checks if a name is a registered function.
     *
     * @param string $name
     * @return boolean
     */
    public function isRegisteredFunction($name)
    {
        return isset($this->_functions[$name]);
    }

    /**
     * Run the CLI script.
     *
     * @return integer The exit value
     */
    final public function run()
    {
        try {
            return $this->_execRun();
        } catch (CliCore\Exception $e) {
            $this->fail($e->getMessage());
        } catch (\Exception $e) {
            $this->getLogger()->error('Uncaught exception');
            $this->getLogger()->error($e->getMessage());
        }

        return self::NOT_OK;
    }

    /**
     * Exec run.
     *
     * @return mixed
     */
    private function _execRun()
    {
        $this->getopts();

        $CamelFilter = new \Zend_Filter_Word_DashToCamelCase();

        // keeps track if any method was successfully called
        $calls = 0;
        // the value that is returned by this function
        $return = 0;

        // see if there is a corresponding function for each parameter, and call it
        foreach (self::$_opts->getOptions() as $option) {
            $camel    = $CamelFilter->filter($option);
            $camel[0] = strtolower($camel[0]); // lcfirst
            $camel    = "_{$camel}";

            if ($this->_isKnownFunction($camel)) {
                try {
                    ++$calls;
                    $return = $this->$camel(self::$_opts->getOption($option));
                } catch (\Zend_Console_Getopt_Exception $e) {
                    $this->_handleGetoptException($e);
                }
            }
        }

        // if we get this far, check for a 'default' function to call
        if (method_exists($this, '_default')) {
            ++$calls;
            $return = $this->_default();
        }
        // if there were no successful function calls, something is wrong, so
        // print usage and return (exit) non-zero
        if ($calls === 0) {
            print($this->getUsage());
            $return = self::NO_ACTION_FOUND;
        }
        return $return;
    }

    /**
     * Run the getopt routine and parse the CLI args.
     *
     * If there are errors then bail with usage info
     *
     * @return void
     */
    public function getopts()
    {
        if (self::$_opts === null) {
            $opts = $this->_getOptArgs();

            try {
                self::$_opts = $this->_getOptFactory($opts);
                self::$_opts->setAliases($this->_getOptAliases());
                self::$_opts->parse();
            } catch (\Zend_Console_Getopt_Exception $e) {
                $this->_handleGetoptException($e);
                // does not return
            }
        }
    }

    /**
     * Add the help option.
     *
     * @return array
     */
    private function _getOptArgs()
    {
        $this->_initDefaultFuncs();
        return $this->_opt_args;
    }

    /**
     * Initialize any default functions.
     */
    protected function _initDefaultFuncs()
    {
        $this->_func('help', 'Print this help message');
        $this->_param('verbose', 'Provide verbose output if possible');
        $this->_param('no-color', 'Suppress colorized output');
        $this->_param(
            'force-color', 'Forces colorized output.  Overrides --no-color'
        );
    }

    /**
     * Shorthand for registerFunction().
     *
     * @param string $opt
     * @param string $description
     */
    protected function _func($opt, $description)
    {
        $this->registerFunction($opt, $description);
    }

    /**
     * Register a function.
     *
     * @param string $opt
     * @param string $description
     */
    public function registerFunction($opt, $description)
    {
        $name = $this->_getCamelFromOpt($opt);

        $this->_functions[$name] = true;
        $this->_addOptArg($opt, $description);
    }

    /**
     * Get the camel-case version of the function name from the opt.
     *
     * @param string $opt
     * @return string
     */
    private function _getCamelFromOpt($opt)
    {
        $Filter = new \Zend_Filter_Word_DashToCamelCase();

        $parts   = explode('|', $opt);      // Split at pipes
        $parts   = explode('=', $parts[0]); // Now split at =
        $dash    = $parts[0];                 // The first part is the name
        $name    = $Filter->filter($dash);  // Convert to camel case
        $name[0] = strtolower($name[0]);    // Lower-case first word
        $name    = '_' .
            $name;               // Add the underscore, since it's protected

        return $name;
    }

    /**
     * Add an option to the opt args array.
     *
     * @param string $opt
     * @param string $description
     */
    private function _addOptArg($opt, $description)
    {
        $this->_opt_args[$opt] = $description;
    }

    /**
     * Shorthand for registerParameter().
     *
     * @param string $opt
     * @param string $description
     */
    protected function _param($opt, $description)
    {
        $this->registerParameter($opt, $description);
    }

    /**
     * Registers a parameter.
     *
     * @param string $opt
     * @param string $description
     */
    public function registerParameter($opt, $description)
    {
        $this->_addOptArg($opt, $description);
    }

    /**
     * Get the options parser.
     *
     * @param array $opts
     * @return \Zend_Console_Getopt
     */
    protected function _getOptFactory(array $opts)
    {
        return new \Zend_Console_Getopt($opts);
    }

    /**
     * Gets aliases for defined options.
     *
     * @return array
     */
    protected function _getOptAliases()
    {
        return [];
    }

    /**
     * Handle opt exceptions by printing the message and the usage message, then
     * exit.
     *
     * @param \Zend_Console_Getopt_Exception $e
     */
    private function _handleGetoptException(\Zend_Console_Getopt_Exception $e)
    {
        print($e->getMessage());
        print("\n\n");
        print($e->getUsageMessage());
        print("\n");
        exit(-1);
    }

    /**
     * Checks if a function exists.
     *
     * @param string $camel_name
     * @return boolean
     */
    private function _isKnownFunction($camel_name)
    {
        if (isset($this->_functions[$camel_name])) {
            return true;
        }

        return false;
    }

    /**
     * Shortcut function to the opts usage message.
     *
     * @return string
     */
    public function getUsage()
    {
        $dont_show = $this->dontDisplayArgs();
        $opts      = $this->_getOptArgs();

        foreach ($dont_show as $key) {
            if (isset($opts[$key])) {
                unset($opts[$key]);
            }
        }

        $tmp_opts = $this->_getOptFactory($opts);
        $tmp_opts->parse();

        return $tmp_opts->getUsageMessage();
    }

    /**
     * Values which should not be displayed in the usage listing.
     *
     * @return array
     */
    public function dontDisplayArgs()
    {
        return [];
    }

    /**
     * Prints a message and fails.
     *
     * @param string|array $msg
     * @param integer      $exit_code
     */
    public function fail($msg, $exit_code = self::NOT_OK)
    {
        $this->info($msg, true, '%R');
        $this->getLogger()->error($msg, ['retval' => $exit_code]);
        exit($exit_code);
    }

    /**
     * Prints translated text for CLI.
     *
     * @param string|array $msg
     * @param boolean      $no_log
     * @param string       $color
     */
    public function info($msg, $no_log = false, $color = null)
    {
        $msg = $this->translate($msg);
        if ($msg !== '') {
            if ($no_log !== true) {
                $this->getLogger()->info($msg);
            }
            $this->echoLine("$msg", $color);
        }
    }

    /**
     * Translate the input.
     * This function really just provides a hook for implementations to translate,
     * here there is no actual translation done.
     *
     * @param string|array $input
     * @return mixed
     */
    public function translate($input)
    {
        return $input;
    }

    /**
     * Print a message to the console.
     *
     * @param string|array $message
     * @param string       $color
     */
    public function echoLine($message, $color = null)
    {
        if (is_array($message)) {
            $message = implode("\n", $message);
        }

        $message = "{$color}{$this->translate( $message )}%n\n";
        $message = $this->_ColorConsole->convert($message);

        if ($this->_isColor() === false) {
            $message = $this->_ColorConsole->strip($message);
        }

        print $message;
    }

    /**
     * Determine if color should be used in output.
     *
     * @return boolean
     */
    protected function _isColor()
    {
        if ($this->_getOption('force-color')) {
            return true;
        }

        if ($this->_getOption('no-color')) {
            return false;
        }

        return posix_isatty(STDOUT);
    }

    /**
     * Retrieves option value from the command line.
     *
     * @param string $flag
     * @param mixed  $default
     * @return mixed
     */
    final protected function _getOption($flag, $default = null)
    {
        if(!self::$_opts) {
            // Help debug this bullshits.
            debug_print_backtrace();
        }

        $opt = self::$_opts->getOption($flag);

        if (in_array($flag, $this->_paths)) {
            $Filter = new PathFilter();
            $opt    = $Filter->filter($opt);
        }

        if ($opt === null) {
            return $default;
        }

        return $opt;
    }

    /**
     * Prints a message to the screen and exits successfully.
     *
     * @param string|array $msg
     */
    public function success($msg = '')
    {
        $this->info($msg, true, '%G');
        exit(self::OK);
    }

    /**
     * Prints a message if the program is being run in verbose mode.
     *
     * @param string|array $message
     * @param string       $color
     */
    public function echoVerbose($message, $color = null)
    {
        if ($this->isVerbose()) {
            $this->echoLine($message, $color);
        }
    }

    /**
     * Checks if the verbose flag was set.
     *
     * @return boolean
     */
    public function isVerbose()
    {
        return $this->_getOption('verbose') === true;
    }

    /**
     * This is just to see if a function is callable from the CLI object's POV.
     *
     * @param mixed $args
     * @return boolean
     */
    public function isCallable($args)
    {
        return is_callable($args);
    }

    // @codingStandardsIgnoreStart

    /**
     * Destructor.
     */
    public function __destruct()
    {
        $this->getLogger()->info('script end');
    }

    /**
     * Short internal version of registerFilePath().
     *
     * @param string $opt
     * @param string $description
     */
    protected function _path($opt, $description)
    {
        $this->registerFilePath($opt, $description);
    }

    /**
     * Registers a parameter that's a file path.
     *
     * @param string $opt
     * @param string $description
     */
    public function registerFilePath($opt, $description)
    {
        $this->registerParameter($opt, $description);

        $parts = explode('|', $opt);      // Split at pipes
        $parts = explode('=', $parts[0]); // Now split at =

        $this->_paths[] = $parts[0];
    }

    /**
     * Makes sure required parameters exist.
     *
     * @param string $func
     * @param string $additional_args
     */
    protected function _requires($func, $additional_args)
    {
        // @codingStandardsIgnoreEnd
        $args  = func_get_args();
        $param = array_shift($args);
        $bail  = false;

        foreach ($args as $arg) {
            if (!$this->_isOptionDefined($arg)) {
                $bail = true;
                if ($func !== 'default') {
                    print("--{$param} option requires --{$arg}\n");
                } else {
                    print("--{$arg} is required\n");
                }
            }
        }

        if ($bail) {
            if ($func === 'default') {
                print($this->getUsage());
            }
            exit(self::PARAM_REQUIRES_CODE);
        }
    }

    /**
     * Check whether the given option name is defined or not.
     *
     * @param string $flag
     * @return boolean
     */
    final protected function _isOptionDefined($flag)
    {
        return self::$_opts->__isset($flag);
    }

    /**
     * Print the usage message and exit.
     */
    protected function _help()
    {
        print($this->getUsage());
        exit(self::OK);
    }

    /**
     * Takes the CLI options and turns them into an array.
     *
     * @param string $option_name
     * @return array
     */
    final protected function _splitOptions($option_name)
    {
        $option = $this->_getOption($option_name);
        $args   = preg_split('/[\s]+/', $option);
        return $args;
    }
}

