<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Matt Parmelee <mparmelee@interworx.com>
 * @date    Dec 14, 2011
 */

/**
 * Factory for CloudFlare configuration.
 *
 * @package    InterWorx
 * @subpackage Payload
 */
class Payload_Factory_SW_Cloudflare
{

    /**
     * Domains payload.
     *
     * @param SiteWorx $SW
     * @param string   $domain
     * @return Payload
     */
    public static function subdomains(SiteWorx $SW, $domain = null)
    {
        $plugin = IW::PluginManager()->getPlugin('cloudflare');
        if ($SW->isParentReseller()) {
            $host_key = $plugin->getDatastoreVar('reseller_host_key');
        } else {
            $host_key = $plugin->getDatastoreVar('host_key');
        }
        $API = new IWorx_CloudflareAPI($host_key);
        $userdata = $plugin->getDatastoreArray('user_key', $SW->getId());
        list( $email, $user_key ) = explode('|', $userdata);
        $domains = self::_getDomains($SW, $domain);

        //set statuses
        foreach ($domains as &$domain) {
            $name   = $domain['subdomain'];
            $parent = $domain['parent'];

            if ($domain['record'] === "points to {$name}.cdn.cloudflare.net") {
                if ($name === "www.{$parent}") {
                    $Record = new Dns_Record_CNAME("host='{$name}'");
                    $alias = $Record->getAlias();
                    $domain['status'] = ( $alias === $parent ) ? 0 : 1;
                } else {
                    $domain['status'] = 1;
                }
            } else {
                $domain['status'] = 0;
            }
        }

        $Payload = new Payload($domains);
        $Payload->setName('cloudflare_domains');
        $Payload->setTitle('Cloudflare DNS Settings');

        $Payload->setColumn(new Payload_Column('type'));

        $Payload->setColumn(new Payload_Column('subdomain'))
        ->setIsSortable(true)
        ->setOrderBy('subdomain')
        ->setLabel('Subdomain');

        $Payload->setColumn(new Payload_Column('record'));

        $Payload->setColumn(new Payload_Column('status'))
        ->addFormatter(new IWorx_Formatter_CloudflareEnabledDisabled('subdomain'))
        ->setLabel('##LG_STATUS##');

        $Payload->setGrouper(new IWorx_Formatter_Grouper('parent'));
        $Payload->setDefaultOrder('parent');

        return $Payload;
    }

    /**
     * Domains payload.
     *
     * @param SiteWorx $SW
     * @return Payload
     */
    public static function domains(SiteWorx $SW)
    {
        $plugin = IW::PluginManager()->getPlugin('cloudflare');
        if ($SW->isParentReseller()) {
            $host_key = $plugin->getDatastoreVar('reseller_host_key');
        } else {
            $host_key = $plugin->getDatastoreVar('host_key');
        }

        $slaves   = $SW->getDomainNameList($type = 'slave');
        $pointers = $SW->getDomainNameList($type = 'pointer');
        $names = array_merge($slaves, $pointers);
        $names[] = $SW->getMasterDomainName();
        $domain_data = [];
        foreach ($names as $dom) {
            $domain_data[] = (object) array( 'domain' => $dom,
                                       'status' => self::_getCFStatus($dom) );
        }

        $Payload = new Payload($domain_data);
        $Payload->setName('cloudflare_settings');
        $Payload->setTitle('Cloudflare Settings');

        /**
         * Returns true if the settings and stats links should be disabled.
         *
         * @param string $value
         * @param object $row
         * @return boolean
         */
        $disabledCFFunc = function ($value, $row) {
            return $row->status === 0;
        };

        $action_data = array( array( 'ctrl'    => 'Ctrl_Siteworx_Cloudflare',
                                 'action'  => 'queryDomainsCommit',
                                 'query'   => 'domain=%%domain%%',
                                 'display' => 'Edit Domains',
                                 'handled' => false,
                                 'class'   => 'iw-cf-dns-edit-link' ),
                          array( 'ctrl'    => 'Ctrl_Siteworx_Cloudflare',
                                 'action'  => 'settings',
                                 'query'   => 'domain=%%domain%%',
                                 'display' => 'Advanced Settings',
                                 'disabledFunc' => $disabledCFFunc,
                                 'handled' => true ),
                          array( 'ctrl'    => 'Ctrl_Siteworx_Cloudflare',
                                 'action'  => 'showStatisticsCommit',
                                 'query'   => 'domain=%%domain%%',
                                 'display' => 'View Statistics',
                                 'disabledFunc' => $disabledCFFunc,
                                 'handled' => true ) );

        $Payload->setColumn(new Payload_Column('action'))
        ->addFormatter(IWorx_Formatter_Action::factory($action_data))
        ->setField('domain')
        ->setLabel('##LG_ACTION##');

        $Payload->setColumn(new Payload_Column('domain'))
        ->setIsSortable(true)
        ->setLabel('Parent Domain');

        $Payload->setColumn(new Payload_Column('status'))
        ->setIsSortable(true)
        ->addFormatter(new IWorx_Formatter_CloudflareEnabledDisabled('domain'))
        ->setLabel('##LG_STATUS##');

        $Payload->refreshOn('settingsCommit');
        return $Payload;
    }

    /**
     * Check the CloudFlare-enabled status of a domain.
     *
     * @param string $domain
     * @return integer
     */
    private static function _getCFStatus($domain)
    {
        $sql = 'SELECT COUNT(*)
            FROM dns_records r,
                 dns_zones z
            WHERE r.zone_id = z.zone_id AND
                  z.domain = ' . q($domain) . ' AND
                  r.type = "CNAME" AND
                  r.target LIKE "%cdn.cloudflare.net"';
        $cnt = IWorxDbUtil::db_string(IW::DB(), $sql);
        if ($cnt > 0) {
            $result = 1;
        } else {
            $result = 0;
        }
        return $result;
    }

    /**
     * Get Siteworx domain data.
     *
     * @param SiteWorx $SW
     * @param string   $domain
     * @return array
     */
    private function _getDomains(SiteWorx $SW, $domain = null)
    {
        $domains = [];
        $names   = [];

        if ($domain === null) {
            $slaves = $SW->getDomainNameList($type = 'slave');
            $pointers = $SW->getDomainNameList($type = 'pointer');

            $names = array_merge($slaves, $pointers);
            $names[] = $SW->getMasterDomainName();
        } else {
            $names[] = $domain;
        }

        foreach ($names as $name) {
            $dns = $SW->getSimpleDnsRecordsFor($name);
            foreach ($dns as $record) {
                if (preg_match('/^ftp\.*/', $record['host'])) {
                    continue;
                }

                if (preg_match('/^cloudflare-resolve-to\.*/', $record['host'])) {
                    continue;
                }
                if ($record['type'] === Dns_Record::CNAME) {
                    $tmp = array( 'type'      => 'CNAME',
                        'subdomain' => $record['host'],
                        'record'    => 'points to ' . $record['target'],
                        'parent'    => $name );
                    $domains[] = $tmp;
                }
            }
        }
        return $domains;
    }
}
